#
# Copyright (C) 2011-2017 Ary Pablo Batista <arypbatista@gmail.com>, Pablo Barenbaum <foones@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

"""Representation of Gobstones boards."""

import random

from ..utils import seq_reversed
from ..i18n import translate as t
from .formats import *
from .basic import BoardFormatException
from ..builtins.types.color import NUM_COLORS, COLORS_BY_NAME

class SelfDestructionException(Exception):
    """Base exception for board-related errors, such as moving the head
    past the board limits, or attempting to take stones when there are
    no stones."""
    pass

def stone_dist():
    """Return a random number of stones, for populating random boards.
    The distribution is taken from the original Fidel's Gobstones/Haskell
    implementation."""
    num = random.randint(0, 999)
    dist = [(650, 0),
            (800, 1),
            (900, 2),
            (950, 3),
            (970, 4)]
    for key, val in dist:
        if num < key:
            return val
    return num - 970 + 4

def rand_side():
    """Return a random length for the board sides."""
    side = random.gauss(9, 3)
    side = max(1, min(side, 200))
    return int(side)

class Cell(object):
    "Represents a Gobstones board cell."

    def __init__(self):
        self.stones = {}

    def randomize(self):
        """Randomizes the cell, filling it with a random number of
        stones for each color."""
        for i in range(NUM_COLORS):
            self.stones[i] = self.stones.get(i, 0) + stone_dist()

    def put(self, coli, count=1):
        """Add `count` stones of the given color index `coli`.
        Note that `coli` is a color index (ord), not an instance of
        Color."""
        self.stones[coli] = self.stones.get(coli, 0) + count

    def take(self, coli):
        """Takes a stone of the given color index `coli`. Note
        that `coli` is a color index (ord), not an instance of
        Color. Raise a SelfDestructionException if
        there are no stones of the given color."""
        cnt = self.stones.get(coli, 0)
        if cnt > 0:
            self.stones[coli] = cnt - 1
        else:
            raise SelfDestructionException(t('Cannot take stones'))

    def set_num_stones(self, coli, count):
        """Set the number of stones of the given color index `coli`
        to `count`. Note that `coli` is a color index (ord), not an instance of
        Color."""
        self.stones[coli] = count

    def num_stones(self, coli):
        """Return the number of stones of the given color index `coli`.
        Note that `coli` is a color index (ord), not an instance of
        Color."""
        return self.stones.get(coli, 0)

    def clone(self):
        "Return a new cell with the same contents as this one."
        copy = Cell()
        copy.clone_from(self)
        return copy

    def clone_from(self, other):
        """Set the contents of this cell to be equal to the contents of
        the other cell."""
        for col, count in other.stones.items():
            self.set_num_stones(col, count)

    def equal_contents(self, other):
        """Return a boolean indicating if this cell has the same contents
        as the other cell."""
        for col, count in self.stones.items():
            if other.stones.get(col, 0) != count:
                return False
        for col, count in other.stones.items():
            if self.stones.get(col, 0) != count:
                return False
        return True

    #### Human-friendly API

    def __repr__(self):
        """Shows the contents of the cell, as a 1x1 board."""
        board = Board((1, 1))
        board[0, 0].clone_from(self)
        return repr(board)

    def __getitem__(self, color_name):
        """Return the number of stones of the given color. For instance:
        >>> cell['Rojo']
        returns the number of red stones in the current cell
        """
        dic = COLORS_BY_NAME
        coli = dic[color_name].ord()
        return self.num_stones(coli)

    def __setitem__(self, color_name, count):
        """Set the number of stones of the given color. For instance:
        >>> cell['Rojo'] = 3
        sets the number of red stones in the current cell to 3
        """
        dic = COLOR_NAME_TO_INDEX_DICT
        coli = dic[color_name]
        self.set_num_stones(coli, count)

    def init_color_properties(cls):
        """Helper method to initialize the cell properties,
        adding class level getters/setters, so that the contents of
        the cell can be modified by means of attributes:
        >>> cell = Cell()
        >>> cell.Rojo = 1
        >>> cell.Rojo
        1
        """
        def init_col(colname, coli):
            "Initialize the accesors for the given color index."

            def pget(self):
                "Color property getter."
                return self.num_stones(coli)

            def pset(self, num):
                "Color property setter."
                self.set_num_stones(coli, num)

            prop = property(fget=pget, fset=pset)
            setattr(Cell, colname, prop)

        for colname, coli in COLORS_BY_NAME.items():
            if not isinstance(colname, str):
                continue
            init_col(colname, coli)

    init_color_properties = classmethod(init_color_properties)

## Initialize color properties for cells
Cell.init_color_properties()
##

class Board(object):
    """Represents a Gobstones board."""

    def __init__(self, size=(1,1)):
        self.size = size
        self.head = (0, 0)
        self.cells = None
        self._clear_board()

        self.changelog_stack = None
        self.changelog = None
        self.clear_changelog()

        self.invalid = False

    def resize(self, width, height):
        "Change the size of the board to width x height."
        self.clone_from(Board((width, height)))

    def go_to_origin(self):
        "Move the head to the origin (0, 0)."
        head_pos = self.head
        self.head = (0, 0)
        self.changelog.append(('move-head', head_pos))

    def go_to_boundary(self, direction):
        "Move the head to specified boundary."
        head_pos = self.head
        width, height = self.size
        y, x = self.head
        delta_y, delta_x = direction.delta()

        if delta_y > 0:
            y = height-1
        elif delta_y < 0:
            y = 0

        if delta_x > 0:
            x = width-1
        elif delta_x < 0:
            x = 0

        self.head = y, x

        self.changelog.append(('move-head', head_pos))

    def _clear_board(self):
        "Clear the contents of the board."
        width, height = self.size
        self.cells = [[Cell() for _ in range(width)] for _ in range(height)]

    def clear_board(self):
        """Clear the contents of the board, saving the original board in
        the changelog, to be able to restore it after returning from a
        Gobstones function."""
        original = self.clone()
        self._clear_board()
        self.changelog.append(('restore', original))

    def set_num_stones(self, y, x, color, count):
        self.cells[y][x].set_num_stones(color.ord(), count)

    def put_stone(self, color):
        """Put a stone of the given color in the current cell."""
        y, x = self.head
        self.cells[y][x].put(color.ord())

        # Save the undo operation in the changelog, for restoring the
        # board after returning from a Gobstones function call.
        self.changelog.append(('take', color))

    def take_stone(self, color):
        """Take a stone of the given color from the current cell."""
        y, x = self.head
        self.cells[y][x].take(color.ord())

        # Save the undo operation in the changelog, for restoring the
        # board after returning from a Gobstones function call.
        self.changelog.append(('put', color))

    def move(self, direction):
        """Move the head to the given direction.
        Raise a SelfDestructionException if the head falls out of the
        board limits."""
        delta_y, delta_x = direction.delta()
        width, height = self.size
        y, x = self.head
        x += delta_x
        y += delta_y
        if 0 <= y and y < height and 0 <= x and x < width:
            self.head = y, x

            # Save the undo operation in the changelog, for restoring the
            # board after returning from a Gobstones function call.
            self.changelog.append(('move', direction.opposite()))
        else:
            raise SelfDestructionException(t('Cannot move'))

    def num_stones(self, color):
        """Return the number of stones of the given color in the current
        cell."""
        y, x = self.head
        return self.cells[y][x].num_stones(color.ord())

    def exist_stones(self, color):
        """Return True iff there are stones of the given color in the
        current cell."""
        y, x = self.head
        return self.cells[y][x].num_stones(color.ord()) > 0

    def can_move(self, direction):
        """Return True iff the head can move in the given direction
        without falling off the board."""
        delta_y, delta_x = direction.delta()
        width, height = self.size
        y, x = self.head
        x += delta_x
        y += delta_y
        return 0 <= y and y < height and 0 <= x and x < width

    def randomize_header(self):
        """Move the head to a random position inside the board."""
        width, height = self.size
        self.head = (
            random.randint(0, height - 1),
            random.randint(0, width - 1))

    def randomize_full(self):
        """Randomize the board size, contents and head position."""
        self.size = rand_side(), rand_side()
        self.randomize_contents()

    def randomize_contents(self):
        """Randomize the board contents and head position.
        Keep the original size."""
        self._clear_board()
        width, height = self.size
        self.randomize_header()
        num_nonempty = random.randint(0, height * width)
        for _ in range(num_nonempty):
            i = random.randint(0, width - 1)
            j = random.randint(0, height - 1)
            self.cells[j][i].randomize()
        self.clear_changelog()

    def randomize(self):
        "Randomize the board."
        self.randomize_full()
        return self

    def dump(self, f, fmt=DefaultFormat, **kwargs):
        "Dump the board to the file handle `f` in the given format."
        if fmt in AvailableFormats:
            AvailableFormats[fmt]().dump(self, f, **kwargs)
        else:
            raise BoardFormatException(
                'Board file format unrecognized: %s' % (fmt,))

    def load(self, f, fmt=DefaultFormat):
        "Load the board from the file handle `f` in the given format."
        if fmt in AvailableFormats:
            AvailableFormats[fmt]().load(self, f)
        else:
            raise BoardFormatException(
                'Board file format unrecognized: %s' % (fmt,))

    def clone(self):
        "Return a copy of this board."
        copy = Board(self.size)
        copy.clone_from(self)
        return copy

    def clone_from(self, other):
        """Set the contents of this board to the contents of the other board.
        Also, clear the changelog, so this becomes the starting point.
        (The cloning cannot be undone)."""
        self._restore_from(other)
        self.clear_changelog()

    def _restore_from(self, other):
        "Set the contents of this board to the contents of the other board."
        self.size = other.size
        self._clear_board()
        assert self.size == other.size
        width, height = other.size
        self.head = other.head
        for x in range(width):
            for y in range(height):
                self.cells[y][x] = other.cells[y][x].clone()

    def __repr__(self):
        gbt_format = AvailableFormats['gbt']()
        out = gbt_format.numbered_contents(self)
        return '\n'.join(out)

    def equal_contents(self, other):
        """Return True iff the contents of the board are equal to the
        contents of the other board."""
        if self.size != other.size:
            return False
        width, height = self.size
        for x in range(width):
            for y in range(height):
                if not self.cells[y][x].equal_contents(other.cells[y][x]):
                    return False
        return True

    #### Functions to deal with the changelog.

    def clear_changelog(self):
        "Clear the changelog."
        self.changelog_stack = []
        self.changelog = []

    def push_state(self):
        """Begin a fresh changelog, pushing the current changelog
        in the changelog stack. Should be called before entering a
        Gobstones function, to log every operation in a fresh changelog,
        to be able to undo them after returning from the function."""
        self.changelog_stack.append(self.changelog)
        self.changelog = []

    def pop_state(self):
        """Use the current changelog to undo the changes made after
        the last invocation to self.push_state. Pop the previous
        changelog from the stack, making it the current one. Should
        be called after returning from a Gobstones function, to undo
        all the effects that might have been caused."""

        def set_head(head):
            "Set the head of the board to `head`."
            self.head = head

        dispatch = {
            'restore': self._restore_from,
            'move-head': set_head,
            'take': self.take_stone,
            'put': self.put_stone,
            'move': self.move,
        }
        changelog_to_undo = self.changelog
        self.changelog = [] # overwrite
        for change in seq_reversed(changelog_to_undo):
            assert change[0] in dispatch
            dispatch[change[0]](change[1])
        self.changelog = self.changelog_stack.pop()

    #### Human-friendly API

    def __getitem__(self, xy):
        x, y = xy
        return self.cells[y][x]

    def goto(self, x, y):
        "Move the head to the position (x, y)."
        width, height = self.size
        assert 0 <= x and x < width
        assert 0 <= y and y < height
        self.head = y, x

def load_board_from(filename):
    """Load the board from the given filename, attempting to
    recognize the format by the file extension."""
    fmt = format_for(filename)
    board = Board((1, 1))
    f = open(filename, 'r')
    board.load(f, fmt)
    f.close()
    return board

def dump_board_to(board, filename, style='verbose'):
    """Dump the board into the given filename, attempting to
    recognize the format by the file extension."""
    fmt = format_for(filename)
    f = open(filename, 'w')
    board.dump(f, fmt, style=style)
    f.close()



# -*- coding: utf-8 -*-
#### Provides translations of all kinds of messages found in the source.



## Spanish
ES = {

# lexer errors
  'Malformed input - unrecognized symbol: "%s"':
    'Entrada mal formada - símbolo no reconocido: "%s"',
  'Malformed input - unrecognized symbol':
    'Entrada mal formada - símbolo no reconocido',
  'Found: %s\nMaybe should be: %s':
    'Se encontró: %s\nSeguramente debería ser: %s',

# smart lexer errors
  'Repeated symbol: %s':
    'Símbolo repetido: %s',
  '%s cannot be followed by %s':
    '%s no puede estar seguido por %s',
  'return is only valid inside functions or the Main procedure':
    'Sólo se admite return dentro de funciones o en el procedimiento Main.',
  'return must be followed by "("':
    'return tiene que estar seguido por "("\n' +
    'Por ejemplo: return (nroBolitas(Rojo))\n' +
    '             return (42, True)',
  'return from function "%s" must return something':
    'El return de la función "%s" tiene que devolver algo\n' +
    'Por ejemplo: return (nroBolitas(Rojo))\n' +
    '             return (42, True)',
  'Program entrypoint should only return variables':
    'El return del punto de entrada del programa sólo puede devolver variables',
  'procedure name should be an uppercase identifier':
    'El nombre de un procedimiento tiene que ser un identificador con mayúscula.\nPor ejemplo: LlenarTablero',
  'function name should be a lowercase identifier':
    'El nombre de una función tiene que ser un identificador con minúscula.\nPor ejemplo: colorCeldaVecina',
  'Unterminated string': 'Cadena sin terminar',
  'Found closing ")" with no matching open paren':
    'Se encontró un ")" pero no había un paréntesis abierto',
  'Found closing "}" with no matching open brace':
    'Se encontró una "}" pero no había una llave abierta',
  'Found open "(" with no matching closing paren':
    'Se encontró un paréntesis "(" pero nunca se cierra',
  'Found open "{" with no matching closing brace':
    'Se encontró una llave "{" pero nunca se cierra',
  'Maybe there is an extra "%s" at %s':
    'Quizás sobra un "%s" en %s',
  'Found end of file but there are open parens yet':
    'Se llegó al final del archivo pero todavía hay paréntesis abiertos',
  'Found end of file but there are open braces yet':
    'Se llegó al final del archivo pero todavía hay llaves abiertas',

# parser errors
  'Syntax error':
    'Error de sintaxis',
  'Found: %s':
    'Se encontró: %s',
  'Found: %s\nExpected: %s':
    'Se encontró: %s\nSe esperaba: %s',
  'Found: %s\nExpected one of the following tokens:':
    'Se encontró: %s\nSe esperaba alguna de las siguientes palabras:',
  'Procedure name "%s" is missing a "("':
    'Falta un "(" después del nombre del procedimiento "%s".',

  'Conflictive rule for: ("%s", "%s")\n':
    'Regla conflictiva para: ("%s", "%s")\n',
  'Will choose first production:\n':
    'Se elige la primera producción:\n',

# smart parser errors
  'Cannot call a function here':
    'No se puede invocar a una función en este punto del programa.',
  'Cannot call a procedure here':
    'No se puede invocar a un procedimiento en este punto del programa.',
  'Premature end of input':
     'Se llegó al final del archivo y había que seguir.\nRevisá si te faltó algo.',
  'Strings can only be used in a THROW_ERROR command':
    'Las cadenas sólo se pueden usar en un comando THROW_ERROR.\nPor ejemplo: THROW_ERROR("rompiste todo")',
  'THROW_ERROR can only accept a string':
    'El comando THROW_ERROR sólo acepta cadenas.\nPor ejemplo: THROW_ERROR("rompiste todo")',

# lint errors
  'Semantic error':
    'Error en el programa',
  'Empty program':
    'El programa está vacío.',
  'The Main procedure should be the last definition':
    'La definición del procedimiento Main debe ser la última del programa.',
  'There should be an entry point (Main procedure or program block).':
    'El programa debería tener un punto de entrada (procedimiento Main o bloque program).',
  # Records
    'Repeated assignment for field "%s".':
        'Asignación repetida para el campo "%s".',
    'the given value has type %s':
        'el valor dado es de tipo %s',
    'Error while building type %s':
        'Error al construir el tipo %s',
  # Variants
    'Variant body should only contain case declarations.':
        'El cuerpo de un tipo variante solo debe contener declaraciones \'case\'.',
  # cannot define
  'Cannot define type "%s"': 'No se puede definir el tipo "%s"',
  'Cannot define procedure "%s"': 'No se puede definir el procedimiento "%s"',
  'Cannot define function "%s"': 'No se puede definir la función "%s"',
  'Cannot define constant "%s"': 'No se puede definir la constante "%s"',
  'Cannot define variable "%s"': 'No se puede definir la variable "%s"',
  'Cannot define parameter "%s"': 'No se puede definir el parámetro "%s"',
  'Cannot define index "%s"': 'No se puede definir el índice "%s"',
  # already defined
  'type "%s" already defined %s': 'El tipo "%s" ya está definido %s',
  'procedure "%s" already defined %s': 'El procedimiento "%s" ya está definido %s',
  'function "%s" already defined %s': 'La función "%s" ya está definida %s',
  'constant "%s" already defined %s': 'La constante "%s" ya está definida %s',
  'variable "%s" already defined %s': 'La variable "%s" ya está definida %s',
  'parameter "%s" already defined %s': 'El parámetro "%s" ya está definido %s',
  'index "%s" already defined %s': 'El índice "%s" ya está definido %s',
  # is too similar
  '"%s" is too similar to type "%s", defined %s': '"%s" se parece al tipo "%s"\nya definido %s',
  '"%s" is too similar to procedure "%s", defined %s': '"%s" se parece al procedimiento "%s"\nya definido %s',
  '"%s" is too similar to function "%s", defined %s': '"%s" se parece a la función "%s"\nya definida %s',
  '"%s" is too similar to constant "%s", defined %s': '"%s" se parece a la constante "%s"\nya definida %s',
  '"%s" is too similar to variable "%s", defined %s': '"%s" se parece a la variable "%s"\nya definida %s',
  '"%s" is too similar to parameter "%s", defined %s': '"%s" se parece al parámetro "%s"\nya definido %s',
  '"%s" is too similar to index "%s", defined %s': '"%s" se parece al índice "%s"\nya definido %s',
  # is not a
  '"%s" is not a procedure': '"%s" no es un procedimiento',
  '"%s" is not a function': '"%s" no es una función',
  '"%s" is not a constant': '"%s" no es una constante',
  '"%s" is not a variable': '"%s" no es una variable',
  '"%s" is not a parameter': '"%s" no es un parámetro',
  '"%s" is a parameter': '"%s" es un parámetro',
  '"%s" is not a index': '"%s" no es un índice',
  '"%s" is an index': '"%s" es un índice',
  # defined at...
  'procedure "%s" defined %s': 'El procedimiento "%s" está definido %s',
  'function "%s" defined %s': 'La función "%s" está definida %s',
  'constant "%s" defined %s': 'La constante "%s" está definida %s',
  'variable "%s" defined %s': 'La variable "%s" está definida %s',
  'parameter "%s" defined %s': 'El parámetro "%s" está definido %s',
  'index "%s" defined %s': 'El índice "%s" está definido %s',
  # is not defined
  'type or callable "%s" is not defined': 'El tipo o rutina %s no está definido',
  'type "%s" is not defined': 'El tipo %s no está definido',
  'procedure "%s" is not defined': 'El procedimiento "%s" no está definido',
  'function "%s" is not defined': 'La función "%s" no está definida',
  'constant "%s" is not defined': 'La constante "%s" no está definida',
  'variable "%s" is not defined': 'La variable "%s" no está definida',
  'parameter "%s" is not defined': 'El parámetro "%s" no está definido',
  'index "%s" is not defined': 'El índice "%s" no está definido',
  #
  'as a built-in': 'por Gobstones',
  'at %s': 'en:\n%s',
  # repeated parameter
  'procedure "%s" has a repeated parameter: "%s"':
    'El procedimiento "%s" tiene un parámetro repetido: "%s"',
  'function "%s" has a repeated parameter: "%s"':
    'La función "%s" tiene un parámetro repetido: "%s"',
  # arity checks
 'Too many arguments for procedure "%s".\nExpected %i (%s), received %i':
   'Demasiados argumentos para el procedimiento "%s".\n' +
   'Esperaba %i argumentos (%s)\nRecibió %i',
 'Too few arguments for procedure "%s".\nExpected %i (%s), received %i':
   'Faltan argumentos para el procedimiento "%s".\n' +
   'Esperaba %i argumentos (%s)\nRecibió %i',
 'Too many arguments for function "%s".\nExpected %i (%s), received %i':
   'Demasiados argumentos para la función "%s".\n' +
   'Esperaba %i argumentos (%s)\nRecibió %i',
 'Too few arguments for function "%s".\nExpected %i (%s), received %i':
   'Faltan argumentos para la función "%s".\n' +
   'Esperaba %i argumentos (%s)\nRecibió %i',
  # return arity
 'Function "%s" returns one value':
   'La función "%s" devuelve un solo valor',
 'Function "%s" returns %i values':
   'La función "%s" devuelve %i valores',
 'One value is expected':
   'Se espera recibir un solo valor',
 '%i values are expected':
   'Se espera recibir %i valores',
  #
  'Repeated variable in assignment: "%s"':
    'La variable "%s" está repetida en la asignación',
  'Literals in a switch should be disjoint':
    'No puede haber literales repetidos en un case',
  'Constructors in a match should be disjoint':
    'No puede haber constructores repetidos en un match',
  '\'match\' expression can only be used with a Variant-type value.':
    'La expresión \'match\' solo puede ser utilizada con valores de tipo variante.',
  #
  'Index of a foreach/repeatWith/repeat cannot be a variable: "%s"':
    'El índice de un foreach/repeatWith/repeat no puede ser una variable: "%s"',
  'Index of a foreach/repeatWith/repeat cannot be a parameter: "%s"':
    'El índice de un foreach/repeatWith/repeat no puede ser un parámetro: "%s"',
  #
  'Cannot modify "%s": %s is immutable':
    'No se puede modificar "%s": %s es inmutable',
  'Cannot modify "%s": index of a foreach/repeatWith/repeat is immutable':
    'No se puede modificar "%s": es el índice de un foreach/repeatWith/repeat',

  'Nested foreach/repeatWith/repeat\'s cannot have the same index: "%s"':
    'foreach/repeatWith/repeat anidados no pueden tener el mismo índice: "%s"',
  #
  'function body of "%s" cannot be empty':
    'El cuerpo de la función "%s" no puede estar vacío\nDebe tener un return',
  'function "%s" should have a return':
    'La función "%s" debería tener un return',
  'procedure "%s" should not have a return':
    'El procedimiento "%s" no debería tener un return',
  'procedure "%s" should not have a procedure variable':
    'El procedimiento "%s" no debería tener una variable de procedimiento',
  'Can not use passing by reference when using the implicit board.':
    'No se puede utilizar pasaje por referencia cuanto se está utilizando el tablero implícito.',
  'procedure "%s" should have a procedure variable':
    'El procedimiento "%s" debería tener una variable de procedimiento',
  'Unknown command: %s': 'Comando no reconocido: %s',
  'Unknown expression: %s': 'Expresión no reconocida: %s',

# modules
  'Module %s cannot be found': 'No se encuentra el módulo %s',
  'Recursive modules': 'Módulos recursivos',
  'function %s was already imported': 'La función "%s" ya fue importada',
  'procedure %s was already imported': 'El procedimiento "%s" ya fue importado',
  'Error parsing module %s': 'Error de sintaxis al cargar el módulo "%s"',
  'Error linting module %s': 'Error al cargar el módulo "%s"',
  'Error typechecking module %s': 'Error de tipos al cargar el módulo "%s"',
  'Error compiling module %s': 'Error de compilación al cargar el módulo "%s"',

# liveness errors
  'Liveness analysis': 'Advertencia en el uso de las variables',
  'Variable "%s" possibly uninitialized':
    'La variable "%s" podría no tener asignado ningún valor\n' +
    'en este punto del programa',
  'Variable "%s" defined but not used':
    'La variable "%s" se asigna pero no se usa más',
  'Variables "(%s)" defined but not used':
    'Las variables "(%s)" se asignan pero ninguna se vuelve a usar',

# type inference errors
  'Type error': 'Error de tipos',
  'Types do not unify: %s %s':
    'No coinciden los tipos:\n%s\n%s',
  '"%s" has type: %s\nRight hand side has type: %s':
    '"%s" tiene tipo: %s\nEl lado derecho tiene tipo: %s',
  'procedure "%s" should take: %s\nBut takes: %s':
     'El procedimiento "%s" debería poder recibir: %s\nPero recibe: %s',
  'function "%s" should take: %s and return: %s\nBut takes: %s and returns: %s':
     'La función "%s" debería poder recibir: %s y devolver: %s\nPero recibe: %s y devuelve: %s',
  'Expression should have type: %s\nBut has type: %s':
    'La expresión debería tener tipo: %s\nPero tiene tipo: %s',
  'Literal should have type: %s\nBut has type: %s':
    'El literal debería tener tipo: %s\nPero tiene tipo: %s',
  'Expression can\'t have type: %s':
    'La expresión no puede tener el tipo: %s',
  'function "%s" is receiving: %s\nBut should receive: %s':
    'La función "%s" está recibiendo: %s\nPero debería recibir: %s',
  'constructor "%s" is receiving: %s\nBut should receive: %s':
    'El constructor "%s" está recibiendo: %s\nPero debería recibir: %s',
  'function "%s" is called as if it returned: %s\nBut returns: %s':
    'La función "%s" se usa como si devolviera: %s\nPero devuelve: %s',
  'procedure "%s" is receiving: %s\nBut should receive: %s':
    'El procedimiento "%s" está recibiendo: %s\nPero debería recibir: %s',
  '"%s" has type: %s\nFunction "%s" returns: %s':
    '"%s" tiene tipo: %s\nPero la función "%s" devuelve: %s',
  '"%s" is not a valid variant case for "%s" type.':
    '"%s" no es un caso válido para el tipo "%s".',

# type syntax errors
  '"%s" is not a type':
    '"%s" no es un tipo',
  'type "%s" expects %u parameters, but receives %u':
    'El tipo "%s" espera %u parámetros, pero está recibiendo %u',

# virtual machine
    'Runtime error':
        'Error en tiempo de ejecución',
    'Runtime type error':
        'Error de tipos en tiempo de ejecución',
    'Program has no entrypoint':
        'El programa no tiene un punto de entrada (procedimiento Main o bloque program)',
    'Uninitialized variable':
        'Variable no inicializada',
    'Identifier "%s" does not exists.':
        'El identificador %s no existe.',
    'Uninitialized variable: "%s"':
        'Variable no inicializada: "%s"',
    'Self destruction:':
        'Autodestrucción:',
    'Identifier refers to a primitive type and can not be projected.':
        'El identificador denota un tipo primitivo y no puede ser proyectado.',
    'Expression has no matching branch.':
        'La expresión del match no coincide con ninguna rama.',

    # Runtime type errors
    '"%s" field has type %s, but type %s was expected.':
        'El campo "%s" tiene tipo %s, pero se esperaba el tipo %s',

  # builtin runtime errors
    'Division by zero': 'División por cero',
    'Negative exponent': 'Potencia con exponente negativo',
    'Empty list': 'Lista vacía',
    '%s was expected': 'Se esperaba un %s',
    '%s type was expected': 'Se esperaba un tipo %s',
    'Function \'%s\' only applicable on %s':
        'Función \'%s\' aplicable solamente en %s',
    'The record doesn\'t have a field named "%s"\nThe available field names for this record are: %s':
        'El registro no posee un campo llamado "%s"\nLos campos disponibles en este registro son: %s',
    'Concatenation between lists with different inner type.':
        'Concatenación entre listas con diferente tipo interno.',
    '"%s" is not a valid field.':
        '"%s" no es un campo válido.',
    'Cannot apply "." operator to "%s"':
        'No se puede aplicar el operador "." a "%s"',
    '"%s" is not indexable.':
        '"%s" no puede inspeccionarse mediante un indice.',
    'The procedure call variable should be a board.':
        'La variable asociada a la llamada de procedimiento debe ser un tablero.',
    'List cursor does not refers to a valid value.':
        'El cursor de la lista no denota ningún valor.',
    '%s constructor expects %s field(s).':
        'El constructor del tipo %s espera recibir %s campo(s)',
    'Passing by reference is disabled when using implicit board.':
        'El pasaje por referencia está deshabilitado cuando se utiliza el modo de tablero implícito.',

  # dynamic type checking
  'Condition should be a boolean':
    'La condición debería ser un valor booleano',
  'Relational operation between values of different types':
    'Operación relacional entre valores de distintos tipos',
  'Relational operation different from equalities between values of %s are not allowed':
    'No se permiten operaciones relacionales distintas a la igualdad entre valores de %s',
  'Arithmetic operation over non-numeric values':
    'Operación aritmética sobre valores no numéricos',
  'Logical operation over non-boolean values':
    'Operación lógica sobre valores no booleanos',
  'The argument to PutStone should be a color':
    'El argumento de Poner debería ser un color',
  'The argument to TakeStone should be a color':
    'El argumento de Sacar debería ser un color',
  'The argument to Move should be a direction':
    'El argumento de Mover debería ser una dirección',
  'The argument to numStones should be a color':
    'El argumento de nroBolitas debería ser un color',
  'The argument to existStones should be a color':
    'El argumento de hayBolitas debería ser un color',
  'The argument to canMove should be a direction':
    'El argumento de puedeMover debería ser una dirección',
  'The argument to opposite should be a direction or an integer':
    'El argumento de opuesto debería ser una dirección o un entero',

# other errors
  'File %s does not exist': 'El archivo "%s" no existe.',

# getter-setter fields
    #'last': 'ultimo',
    #'head': 'cabeza',
    #'current': 'actual',
    #'tail': 'cola',
    #'init': 'comienzo',

# token names
  'a number': 'un número',
  'a string': 'una cadena',
  'a lowercase identifier': 'un identificador con minúscula',
  'an uppercase identifier': 'un identificador con mayúscula',
  'end of input': 'final del archivo',
  'beggining of input': 'comienzo del archivo',

# AST names
  'function definition': 'definición de función',
  'procedure definition': 'definición de procedimiento',
  'procedure call': 'invocación a procedimiento',
  'function call': 'invocación a función',
  'variable name': 'nombre de variable',
  'variable assignment': 'asignación de variable',
  'relational operator': 'operador relacional',
  'list operator': 'operador de listas',
  'arithmetic operator': 'operador aritmético',
  'logical operator': 'operador lógico',
  'import clause': 'cláusula from..import',

# builtin procedures
  'PutStone': 'Poner',
  'PutStoneN': 'PonerN',
  'TakeStone': 'Sacar',
  'Move': 'Mover',
  'GoToOrigin': 'IrAlOrigen',
  'GoToBoundary': 'IrAlBorde',
  'ClearBoard': 'VaciarTablero',

# list builtin rutines
  'Next': 'Avanzar',
  'GoToStart': 'IrAlInicio',
  'hasNext': 'hayMas',
  'AddFirst': 'AgregarAdelante',
  'Append': 'AgregarAtras',
  'DropFirst': 'SacarPrimero',
  'DropLast': 'SacarUltimo',

# builtin functions
  'numStones': 'nroBolitas',
  'existStones': 'hayBolitas',
  'canMove': 'puedeMover',
  'minBool': 'minBool',
  'maxBool': 'maxBool',
  'minDir': 'minDir',
  'maxDir': 'maxDir',
  'minColor': 'minColor',
  'maxColor': 'maxColor',
  'next': 'siguiente',
  'prev': 'previo',
  'opposite': 'opuesto',

# builtin constants
  'True': 'True',
  'False': 'False',

  'North': 'Norte',
  'South': 'Sur',
  'East': 'Este',
  'West': 'Oeste',

  'Color0': 'Azul',
  'Color1': 'Negro',
  'Color2': 'Rojo',
  'Color3': 'Verde',

# types
  'Color': 'Color',
  'Dir': 'Dir',
  'Bool': 'Bool',
  'Int': 'Int',
  'List': 'List',
  'Array': 'Arreglo',

# boards
  'Cannot show board, max number allowed is %s':
    'No se puede mostrar el tablero.\nEl máximo número permitido es %s',
  'Malformed board':
    'Archivo de tablero mal formado',
  'Loading of html boards not supported':
    'No se pueden cargar tableros en formato HTML',
  'Cannot take stones':
    'No se pueden sacar bolitas',
  'Cannot take stones of color %s':
    'No se puede sacar una bolita de color: %s\nNo hay bolitas de ese color',
  'Cannot move':
    'No se puede mover el cabezal\nLa posición cae afuera del tablero',
  'Cannot move to %s':
    'No se puede mover el cabezal en dirección: %s\nLa posición cae afuera del tablero',
  'Invalid direction':
    'Dirección inválida',

# misc
  'At': 'En',
  'At:': 'En:',
  'Locals:': 'Valores de los identificadores:',
  'to': 'hasta',
  'near': 'cerca de',
  'after': 'después de',
  'line': 'línea',
  'column': 'columna',
  'Prelude': 'Biblioteca',

# misc2
  'atomic values': 'valores atómicos',
  'atomic value': 'valor atómico',
  'list type values': 'valores de tipo lista',
  'list type value': 'valor de tipo lista',
  'string type value': 'valor de tipo cadena',

## command line
  'I18N_gbs_usage':
'''Uso: <PROG> entrada.gbs [opciones]
Opciones:
  [--from] tablero.{gbb,gbt,tex}  Ejecuta el programa en el tablero dado
  --to tablero.{gbb,gbt,tex}      Guarda el resultado en el tablero dado
  --size <ancho> <alto>           Tamaño del tablero generado
  --pprint                        Imprime el código fuente
  --print-ast                     Imprime el árbol sintáctico
  --print-asm                     Imprime el código de la máquina virtual
  --asm salida.gbo                Genera código para la máquina virtual
  --lint {lax,strict}             Nivel de la etapa de lint (default: strict)
  --no-typecheck                  No hace inferencia de tipos
  --no-liveness                   No hace análisis de variables vivas
  --no-print-board                No muestra el resultado por pantalla
  --no-print-retvals              No muestra los valores de retorno
  --style <estilo1,...,estiloN>   Formato para la salida del tablero
          compact                   .gbb,.gbbo: Formato compacto
          [no-]head                 .fig: [No] mostrar el cabezal
          [no-]labels               .fig: [No] mostrar etiquetas de filas y columnas
          [no-]colors               .fig: [No] usar colores
          [no-]color-names          .fig: [No] mostrar nombres de colores
  --jit                           Habilitar compilación Just in Time
  --print-jit                     Mostrar instrucciones del JIT
  --print-native                  Mostrar código nativo obtenido por el JIT
  --license                       Muestra la licencia del programa
Otros usos:
  <PROG> tablero_in.{gbb,gbt,tex} [--to tablero_out.{gbb,gbt,tex,html,fig}]
  <PROG> entrada.gbo [opciones]
''',
  'Wrote %s': 'Se escribió el archivo %s',

## GUI

  '&File': '&Archivo',
  '&New': '&Nuevo',
  '&Open': '&Abrir',
  '&Save': '&Guardar',
  'Save &as...': 'Guardar &como...',
  '&Quit': '&Salir',

  '&Edit': '&Editar',
  '&Undo': '&Deshacer',
  '&Redo': '&Rehacer',
  'Cu&t': 'Cor&tar',
  '&Copy': '&Copiar',
  '&Paste': '&Pegar',
  'Select &All': 'Seleccion&ar Todo',
  '&Find': '&Buscar',
  '&Indent region': '&Identar',
  '&Dedent region': '&Desidentar',
  '&Font family': '&Fuente',
  'Font &size': '&Tamaño de la fuente',
  '(&More...)': '(&Más...)',

  '&Gobstones': '&Gobstones',
  '&Run': 'E&jecutar',
  '&Check': '&Chequear',

  '&Board': '&Tablero',
  '&Empty board': '&Vaciar tablero',
  '&Random board': 'Modo aleato&rio',
  '&Random board size': 'Tamaño aleator&io',
  '&Change size': 'Cambiar &tamaño',
  '&Load board': '&Cargar tablero',
  '&Save board': '&Guardar tablero',
  'E&xport TeX': 'E&xportar TeX',
  'I&mport TeX': 'I&mportar TeX',

  '&Help': 'A&yuda',
  '&Version': '&Versión',
  '&License': '&Licencia',

  'File not saved': 'Archivo no guardado',
  'File has not been saved.\nSave it now?':
    'El archivo no fue guardado.\n¿Querés guardarlo ahora?',
  'Source must be saved': 'Hay que guardar el archivo',
  'Source must be saved.\nOk to save?':
    'Hay que guardar el archivo\n¿Querés guardarlo ahora?',

  'Open file': 'Abrir archivo',
  'Save file': 'Guardar archivo',
  'Gobstones source': 'Programa Gobstones',
  'Gobstones board': 'Tablero Gobstones',
  'Gobstones board (old format)': 'Tablero Gobstones (formato antiguo)',
  'TeX/LaTeX file': 'Archivo TeX/LaTeX',
  'HTML file': 'Archivo HTML',
  'All files': 'Todos los archivos',

  'Change board size': 'Cambiar tamaño del tablero',
  'Width:': 'Ancho',
  'Height:': 'Alto',
  'Invalid board size': 'Tamaño de tablero inválido.\nEl ancho y el alto deben ser números entre 1 y 200.',

  'Untitled': 'Sin título',

  'All checks ok': 'El programa está bien formado.\n',

  'Stop': 'Parar',
  'Execution interrupted by the user': 'La ejecución del programa fue interrumpida por el usuario.\nEs posible que el programa se haya colgado debido a un while que no termina.',

  'Gobstones version': 'Gobspy versión',

# Board viewer
  'Initial board': 'Tablero inicial',
  'Final board': 'Tablero final',
  'Board viewer': 'Visor de tableros',
  'There is no board to save': 'No hay tablero para guardar',


# Log messages
  'Parsing.': 'Haciendo análisis sintáctico.',
  'Exploding program macros.': 'Explotando macros del programa.',
  'Performing semantic checks.': 'Haciendo análisis semántico.',
  'Compiling.': 'Compilando.',
  'Starting program execution.': 'Ejecutando el programa.',
  'Program execution finished.': 'Ejecución finalizada.',

# Judge
  '&Problems': 'Pro&blemas',
  '&Solve problem': '&Resolver problema',
  'Solve problem': 'Resolver problema',
  'Problem name:': 'Nombre del problema:',
  'There is no valid @test directive':
    'La herramienta de testeo requiere una directiva @test. Algo como:\n' + \
    '    // @test\n'
    '    procedure PonerVerde()\n',
  'Too many @test directives':
    'Debe haber una sola ocurrencia de la directiva @test.',
  'Problem "%s" does not exist': 'El problema "%s" no existe',
  'Program solving problem "%s" was accepted by judge':
    'La resolución del problema "%s" fue aceptada.\n' + \
    'Pero cuidado:\n' + \
    '"Hacer tests puede mostrar la presencia de errores,\n    nunca su ausencia".\n' + \
    '      [Edsger W. Dijkstra]',
  'Program failed on test case "%s". Debug program and retry.':
    'La resolución del problema no funciona\npara el test "%s".\nCorregí el programa y volvé a probar.',
  'Runtime error. Debug program and retry.':
    'La resolución del problema tiene algún error.\nCorregí el programa y volvé a probar.',
  'Program failed on some test cases. Debug program and retry.':
    'La resolución del problema no funciona\npara alguno de los tests.\nCorregí el programa y volvé a probar.',

# Judge reports
  'Test case "%s" failed': 'Test "%s"',
  'Test case': 'Test',
  'Source code': 'Código fuente',
  'Expected result': 'Esperado',
  'Obtained result': 'Obtenido',
  'Expected final board': 'Tablero final esperado',
  'Obtained final board': 'Tablero final obtenido',
  'Returned variables': 'Variables devueltas',
  'No returned variables': 'Ninguna',
  'No result': 'No se obtuvo un resultado',

# Test driver
  'Starting problem %s': 'Problema %s.',
  'Starting test case %s.': 'Probando tablero inicial: %s',
  'Checking fingerprints': 'Comparando los resultados',

# Version
  'I18N_gbs_version(%s)': 'Gobspy versión %s\nIntroducción a la programación\nUniversidad Nacional de Quilmes, 2011-2017\n',
}

# English
EN = {

  '"%s" is not a index': '"%s" is not an index',

## command line
  'I18N_gbs_usage':
'''Usage: <PROG> source.gbs [options]
Options:
  [--from] board.{gbb,gbt,tex}  Run the program in the given board file
  --to board.{gbb,gbt,tex}      Save the result in the given board file
  --size <width> <height>       Size of the input board when randomized
  --pprint                      Pretty print source code
  --print-ast                   Print the abstract syntax tree
  --print-asm                   Print the code for the virtual machine
  --asm output.gbo              Generate virtual machine code
  --lint {lax,strict}           Strictness of the lint stage (default: strict)
  --no-typecheck                Don't do type inference
  --no-liveness                 Don't do live variable analysis
  --no-print-board              Don't output the result
  --no-print-retvals            Don't output the return values
  --compact                     Use compact format for dumping .gbb and .gbo files
  --jit                         Enable Just in Time compiler
  --print-jit                   Print JIT instructions
  --print-native                Print JIT native code

Other usages:
  <PROG> board_in.{gbb,gbt,tex} [--to board_out.{gbb,gbt,tex,html,fig}]
  <PROG> input.gbo [options]
''',

# builtin constants
  'Color0': 'Cyan',
  'Color1': 'Black',
  'Color2': 'Red',
  'Color3': 'Violet',

# Version
  'I18N_gbs_version(%s)': 'Gobspy version %s\n',

}

Languages = {
    'es' : ES,
    'en' : EN
}

Language = Languages['es']

def set_language(value):
    global Language, Languages
    if value in Languages.keys():
        Language = Languages[value]
    else:
        Language = Languages['en']

def translate(s):
  # return the original string if no translation is found
  return Language.get(s, s)

Token_type_descriptions = {
  'EOF': translate('end of input'),
  'BOF': translate('beggining of input'),
  'num': translate('a number'),
  'string': translate('a string'),
  'lowerid': translate('a lowercase identifier'),
  'upperid': translate('an uppercase identifier'),
}

AST_type_descriptions = {
  'function': translate('function definition'),
  'procedure': translate('procedure definition'),
  'procCall': translate('procedure call'),
  'assignVarName': translate('variable assignment'),
  'varName': translate('variable name'),
  'funcCall': translate('function call'),
  'and': translate('logical operator'),
  'or': translate('logical operator'),
  'not': translate('logical operator'),
  'relop': translate('relational operator'),
  'listop': translate('list operator'),
  'addsub': translate('arithmetic operator'),
  'mul': translate('arithmetic operator'),
  'pow': translate('arithmetic operator'),
  'divmod': translate('arithmetic operator'),
  'unaryMinus': translate('arithmetic operator'),
  'import': translate('import clause'),
}



from .utils import indent

class BoardFormatException(Exception):
    pass

class SourceException(Exception):

    def __init__(self, message, area=None):
        self.message = message
        self.area = area

    def __repr__(self):
        s = ''
        if self.area:
            s += '\n%s\n' % (self.area,)
        s += '%s\n' % (indent(self.message),)
        return s

    def error_type(self):
        return 'Error'


class StaticException(SourceException):
    pass


class DynamicException(SourceException):
    pass



#
# Copyright (C) 2011-2017 Ary Pablo Batista <arypbatista@gmail.com>, Pablo Barenbaum <foones@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

from ..utils import *
from ..i18n import translate as t
from ..builtins.types.color import NUM_COLORS, COLORS_BY_NAME, COLORS, COLOR_NAMES, Color
from .basic import BoardFormatException, BoardFormat

class HtmlBoardFormat(BoardFormat):
  # size can be:
  #   - the string 'relative'
  #   - an integer (side of a cell in pixels)
  def gbs_board_style(self, size='relative', board_size=(9, 9)):
    style = """
table.gbs_board {
  border: black solid 1px;
  border-spacing: 0;
  border-collapse: collapse;
  font-family: Verdana, helvetica;
  font-size: $FONT_SIZE;
}
.gbs_board td {
  margin: 0;
  padding: 2px;
  border: solid #a52a2a 1px;
  background: #ddd;
  width: $WIDTH;
  height: $HEIGHT;
}
.gbs_board td.gh { /* position of the header in the board */
  margin: 0;
  padding: 2px;
  border: dashed #440 3px;
  background: #dd8;
  width: $WIDTH;
  height: $HEIGHT;
}
.gbs_board td.lv { /* labels at the side */
  text-align: center;
  vertical-align: middle;
  border-style: none;
  background: #fff;
  width: $HALF_WIDTH;
}
.gbs_board td.lh { /* labels at the top / bottom */
  text-align: center;
  vertical-align: middle;
  border-style: none;
  background: #fff;
  height: $HALF_HEIGHT;
}
.gbs_board td.lx { /* corner */
  border-style: none;
  background: #fff;
  width: $HALF_WIDTH;
  height: $HALF_HEIGHT;
}
.gbs_board table.gc { /* cell table */
  border: none;
  border-spacing: 2px;
  border-collapse: separate;
}
.gbs_board .gc td {
  border: none;
  width: $HALF_WIDTH;
  height: $HALF_HEIGHT;
  text-align: center;
  vertical-align: middle;
  color: black;
  font-family: Verdana, helvetica;
  font-size: $FONT_SIZE;
}
.gbs_board td.A { background: #88f; border: solid 1px #008; }
.gbs_board td.N { background: #aaa; border: solid 1px #222; }
.gbs_board td.R { background: #f88; border: solid 1px #800; }
.gbs_board td.V { background: #8f8; border: solid 1px #080; }
.gbs_board td.O { background: none; } /* empty */
    """
    font_size = '9pt'
    if size == 'relative':
      tot_width = 100.0
      tot_height = 100.0
      unit = '%'

      w, h = board_size
      half_width = int(tot_width / (2 * (w + 1)))
      half_height = int(tot_height / (2 * (w + 1)))
      if half_width < 5:
        half_width = 5
      if half_height < 5:
        half_height = 5

    else:
      half_width = size / 2
      half_height = size / 2
      if half_width < 5:
        half_width = 5
      if half_height < 5:
        half_height = 5
      unit = 'px'

    width = '%s' % (2 * half_width,)
    height = '%s' % (2 * half_height,)
    half_width = '%s' % (half_width,)
    half_height = '%s' % (half_height,)
    style = style.replace('$WIDTH', width + unit)
    style = style.replace('$HEIGHT', height + unit)
    style = style.replace('$HALF_WIDTH', half_width + unit)
    style = style.replace('$HALF_HEIGHT', half_height + unit)
    style = style.replace('$FONT_SIZE', font_size)
    return style

  def render(self, board, draw_head=True):
    w, h = board.size
    out = StringIO()

    def row_titles():
      out.write('<tr>')
      out.write('<td class="lx">&nbsp;</td>')
      for x in range(w):
        out.write('<td class="lh">%u</td>' % (x,))
      out.write('<td class="lx">&nbsp;</td>')
      out.write('</tr>\n')

    out.write('<table class="gbs_board" cellpadding="0" cellspacing="0">\n')
    row_titles()
    for y in seq_reversed(range(h)):
      out.write('  <tr>\n')
      out.write('    <td class="lv">%u</td>\n' % (y,))
      for x in range(w):
        def cell_for(coli):
          cant = board.cells[y][x].num_stones(coli)
          if cant == 0: return '<td class="O">&nbsp;</td>'
          col = Color(coli).name()
          return '<td class="%s">%i</td>' % (col[0], cant)

        if board.head == (y, x) and draw_head:
          out.write('    <td class="gh">\n')
        else:
          out.write('    <td>\n')

        out.write('      <table class="gc">\n')
        out.write('        <tr>%s%s</tr>\n' % (cell_for(1), cell_for(0)))
        out.write('        <tr>%s%s</tr>\n' % (cell_for(2), cell_for(3)))
        out.write('      </table>\n')

        out.write('    </td>\n')
      out.write('    <td class="lv">%u</td>\n' % (y,))
      out.write('  </tr>\n')
    row_titles()
    out.write('</table>\n')
    res = out.getvalue()
    out.close()
    return res

  def dump(self, board, f, **kwargs):
    f.write('<style type="text/css">')
    #f.write(self.gbs_board_style(size=50, board_size=board.size))
    f.write(self.gbs_board_style(size=30, board_size=board.size))
    f.write('</style>\n\n')
    f.write(self.render(board))

  def load(self, board, f):
    raise BoardFormatException(t('Loading of html boards not supported'))



from .board import Board
from traceback import format_stack, extract_stack, format_list

class State:

    def __init__(self):
        self.init()

    def backtrace(self, message=''):
        return message + '\n' + ''.join(format_list(extract_stack()[:-2]))

    def push(self):
        self.board.push_state()

    def pop(self):
        self.board.pop_state()

    def init(self):
        self.board = Board().randomize()

state = State()



#
# Copyright (C) 2011-2017 Ary Pablo Batista <arypbatista@gmail.com>, Pablo Barenbaum <foones@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import re

from ..utils import *
from ..i18n import translate as t
from ..builtins.types.color import NUM_COLORS, I18N_COLORS_BY_NAME, COLORS, I18N_COLOR_NAMES, Color
from .basic import BoardFormatException, BoardFormat
from math import log10, pow

def prefix_notation(number):
    prefixes = {
        3: "K",
        6: "M",
        9: "G",
        12: "P",
    }
    target_power = int(log10(number))
    surplus = int(number / pow(10, target_power))
    return str(surplus) + prefixes[target_power]

class GbtBoardFormat(BoardFormat):
  "Classic GBT board format inherited from the Haskell implementation."

  def __init__(self, cell_w=9, cell_h=3, max_num_len=3):
    # parameters for ascii output
    self.Cell_w = cell_w
    self.Cell_h = cell_h
    self.Max_num_len = max_num_len

  def load(self, board, f):
    head = None
    contents = list(self._parse_board(f))
    for evt in contents:
      if evt[0] == 'SIZE':
        board.size = evt[2], evt[1]
    w, h = board.size
    for evt in contents:
      if evt[0] == 'HEAD' and evt[1] is not None and evt[2] is not None:
        head = h - evt[1] - 1, evt[2]
    if head is None:
      board.randomize_header()
    else:
      board.head = head
    board._clear_board()
    for evt in contents:
      if evt[0] == 'STONE':
        description = evt[3]
        for d in self._parse_description(description):
          self._put_from_description(board.cells[h - evt[1] - 1][evt[2]], d)
    board.clear_changelog()

  def dump(self, board, f, **kwargs):
    f.write('[\n')
    rows = self.numbered_contents(board)
    i = 0
    for row in rows:
      if i == len(rows) - 1:
        sep = ''
      else:
        sep = ','
      f.write('"%s"%s\n' % (row, sep))
      i += 1
    f.write(']\n')

  def numbered_contents(self, board):
    w, h = board.size
    res = []
    out = self._board_contents(board)
    # vertical numbers
    row_id = h
    i = 0
    for row in out:
      if i % (self.Cell_h + 1) == 2:
        row_id -= 1
        res.append('  %i ' % (row_id,) + ''.join(row) + ' %i ' % (row_id,))
      else:
        res.append('    ' + ''.join(row) + '   ')
      i += 1
    # horizontal numbers
    rf = self.Cell_w // 2
    lf = self.Cell_w - rf
    horiz = ''.join([' ' * lf + '%i' % (i,) + ' ' * rf for i in range(w)])
    horiz = '    ' + horiz + '    '
    res.insert(0, horiz)
    res.append(horiz + '  ')
    return res

  def _board_contents(self, board):
    w, h = board.size
    gw = (self.Cell_w + 1) * w + 1
    gh = (self.Cell_h + 1) * h + 1
    out = [[' ' for i in range(gw)] for j in range(gh)]
    # row borders
    for x in range(w):
      for y in range(h + 1):
        for i in range((self.Cell_w + 1) * x + 1, (self.Cell_w + 1) * (x + 1)):
          out[(self.Cell_h + 1) * y][i] = '-'
    # column borders
    for x in range(w + 1):
      for y in range(h):
        for i in range((self.Cell_h + 1) * y + 1, (self.Cell_h + 1) * (y + 1)):
          out[i][(self.Cell_w + 1) * x] = '|'
    # corners
    for x in range(w + 1):
      for y in range(h + 1):
        out[(self.Cell_h + 1) * y][(self.Cell_w + 1) * x] = '+'
    # head
    y0, x0 = board.head
    y0 = h - y0 - 1
    for y in range(y0, y0 + 2):
      for i in range((self.Cell_w + 1) * x0 + 1, (self.Cell_w + 1) * (x0 + 1)):
        out[(self.Cell_h + 1) * y][i] = 'X'
    for x in range(x0, x0 + 2):
      for i in range((self.Cell_h + 1) * y0 + 1, (self.Cell_h + 1) * (y0 + 1)):
        out[i][(self.Cell_w + 1) * x] = 'X'
    # contents
    for x in range(w):
      for y in range(h):
        cell = self._cell_contents(board.cells[h - y - 1][x])
        for i in range(self.Cell_w):
          for j in range(self.Cell_h):
            out[(self.Cell_h + 1) * y + j + 1][(self.Cell_w + 1) * x + i + 1] = cell[j][i]
    return out

  def _cell_contents(self, cell):
    w, h = self.Cell_w, self.Cell_h
    out = [[' ' for i in range(w)] for j in range(h)]
    for col in range(NUM_COLORS):
      count = cell.num_stones(col)
      if count == 0: continue
      if col < 2:
        y = 0
      else:
        y = self.Cell_h - 1
      if col % 2 == 0:
        x = self.Cell_w - 2
      else:
        x = self.Max_num_len
      out[y][x] = Color(col).i18n_name()[0]
      scount = str(count)
      if len(scount) > self.Max_num_len:
        scount = prefix_notation(count)
      for c in seq_reversed(scount):
        x -= 1
        out[y][x] = c
    return out

  def _parse_description(self, description):
    s = ''
    i = 0
    while i < len(description):
      s += description[i]
      if description[i] not in '0123456789':
        if s != '': yield s
        s = ''
      i += 1
    if s != '': yield s

  def _parse_board(self, f):
    width = None
    height = None
    head_row = None
    head_col = None
    # [
    l = f.readline().strip(' \t\r\n')
    if l != '[':
      raise BoardFormatException(t('Malformed board'))
    def is_separator(l):
      l = re.sub('[X-]', '', l)
      return l == '+' * (width + 1)
    def head_starts_at_row(l):
      return 'X' in l
    def number_header(l):
      l = re.sub(' +', ',', l).split(',')
      if l != [str(i) for i in range(len(l))]:
        raise BoardFormatException(t('Malformed board'))
      return l
    # number header
    l = f.readline().strip(' \t\r\n",')
    width = len(number_header(l))
    # first horizontal line
    l = f.readline().strip(' \t\r\n",')
    if not is_separator(l):
      raise BoardFormatException(t('Malformed board'))
    row = 0
    while True:
      # read a board row, consisting of many file lines
      while True:
        l = f.readline()
        if l == '':
          raise BoardFormatException(t('Malformed board'))
        l1 = l.strip(' \t\r\n",0123456789')
        if is_separator(l1):
          if head_starts_at_row(l1):
            head_row = row
          break
        elif '|' not in l1:
          number_header(l.strip(' \t\r\n"')) # check
          l = f.readline().strip(' \t\r\n')
          if l != ']':
            raise BoardFormatException(t('Malformed board'))
          yield 'HEAD', head_row, head_col
          yield 'SIZE', row, width
          return
        else:
          l2 = re.sub('[^|X]', '', l1)
          if 'X' in l2:
            head_col = l2.index('X')
          line_contents = l1.replace('X', '|').split('|')
          if line_contents[0] != '' or line_contents[-1] != '':
            raise BoardFormatException(t('Malformed board'))
          line_contents = line_contents[1:-1]
          col = 0
          for cell in line_contents:
            for elem in cell.split(' '):
              if elem == '': continue
              yield 'STONE', row, col, elem
            col += 1
      row += 1

  def _put_from_description(self, cell, description):
    coli = 0
    for cn in I18N_COLOR_NAMES:
      if description[-1].lower() == cn[0].lower():
        count = description[:-1]
        for l in count:
          if l not in '0123456789':
            raise BoardFormatException(t('Malformed board'))
        cell.put(coli, int(count))
        return
      coli += 1
    raise BoardFormatException(t('Malformed board'))



#
# Copyright (C) 2011-2017 Ary Pablo Batista <arypbatista@gmail.com>, Pablo Barenbaum <foones@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import re

from ..utils import *
from ..i18n import translate as t
from ..builtins.types.color import NUM_COLORS, I18N_COLORS_BY_NAME, Color
from .basic import BoardFormatException, BoardFormat

def is_numeric(x):
  return re.match('^[0-9]+$', x)

class GbbBoardFormat(BoardFormat):
  "Simple human-friendly board format."

  def dump(self, board, f, style='verbose', **kwargs):
    if style == 'verbose':
      self.dump_with_translator(board, f, self.verbose_translator)
    elif style == 'compact':
      self.dump_with_translator(board, f, self.compact_translator)
    else:
      assert False

  def verbose_translator(self, s):
    return s

  def compact_translator(self, s):
    _dict = {
      'cell': 'o',
      'size': 's',
      'head': 'x',
    }
    if s in _dict:
      return _dict[s]
    else:
      return s[0]

  def dump_with_translator(self, board, f, translate):
    w, h = board.size
    f.write('GBB/1.0\r\n')
    f.write('%s %i %i\r\n' % (translate('size'), w, h))
    for x in range(w):
      for y in range(h):
        cell = []
        for coli in range(4):
          cant = board.cells[y][x].num_stones(coli)
          if cant == 0: continue
          col = Color(coli).i18n_name()
          cell.append('%s %i' % (translate(col), cant))
        if cell == []: continue
        f.write('%s %i %i %s\r\n' % (translate('cell'), x, y, ' '.join(cell)))
    y, x = board.head
    f.write('%s %i %i\r\n' % (translate('head'), x, y))
    f.write('%%\r\n')

  def load(self, board, f):

    orig = ['']

    f_lines = read_stripped_lines(f)

    def next_line():
      if f_lines == []:
        return False
      else:
        orig[0] = f_lines.pop(0)
        return orig[0].split(' ')

    def fail(msg):
      raise BoardFormatException(t('Malformed gbb board') + '\n' +
                                 '  ' + t('Near line:') + ' "' + orig[0].strip('\r\n') + '"\n' +
                                 '  ' + msg)

    # GBB header
    l = next_line()
    if l != ['GBB/1.0']:
      fail(t('Expected header line "GBB/1.0"'))

    l = next_line()
    if len(l) == 3 and l[0] in ['size', 's'] and is_numeric(l[1]) and is_numeric(l[2]):
      width, height = board.size = int(l[1]), int(l[2])
    else:
      fail(t('Expected board size line "size <width> <height>"'))

    board.randomize_header()
    board._clear_board()

    color_to_index = I18N_COLORS_BY_NAME

    visited_cells = {}

    while True:
      l = next_line()
      if not l: break
      if l[0] in ['head', 'x']:
        if len(l) == 3 and is_numeric(l[1]) and is_numeric(l[2]):
          x, y = int(l[1]), int(l[2])
          if x >= width or y >= height:
            fail(t('"head %u %u" out of bounds') % (x, y))
          board.head = y, x
        else:
          fail(t('Expected head line "head <x> <y>"'))
      elif l[0] in ['cell', 'o']:
        if len(l) >= 3 and is_numeric(l[1]) and is_numeric(l[2]):
          x, y = int(l[1]), int(l[2])
          if (x, y) in visited_cells:
            fail(t('Cell (%u,%u) is repeated.') % (x, y))
          visited_cells[(x, y)] = True
          count_col = {}
          rest = l[3:]
          while rest != []:
            color_name = rest[0].capitalize()
            coli = None
            for k in color_to_index.keys():
                if color_name == k or k[0] == color_name[0]:
                    coli = color_to_index[k]

            if len(rest) >= 2 and not coli is None and is_numeric(rest[1]):
              rest.pop(0)
              count_col[coli] = count_col.get(coli, 0) + int(rest.pop(0))
            else:
              fail(t('Expected cell line "cell <x> <y> [<color> <num>]*"'))
          for k, v in count_col.items():
            board.set_num_stones(y, x, k, v)
        else:
          fail(t('Expected cell line "cell <x> <y> [<color> <num>]*"'))
      else:
        fail(t('Expected head line "head <x> <y>" or cell line "cell <x> <y> [<color> <num>]*"'))



from .color import *
from .direction import *

Color = color.Color
for name, color in color.COLORS_BY_NAME.items():
    code = '%s = color' % name
    exec(code)

Direction = direction.Direction
for name, direction in direction.DIRECTIONS_BY_NAME.items():
    code = '%s = direction' % name
    exec(code)

#types = locals()



import sys
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO
from .builtins import builtins
from .state import state
from .preprocessor import preprocess
from .board import Board
from traceback import format_exc

class GobspyResult:

    def __init__(self, pycode, out, final_board, error=''):
        self.final_board = final_board
        self.out = out
        self.result = []
        self.pycode = pycode
        self.error = error

    def failed(self):
        return self.final_board is None

class Gobspy:

    def run(self, text, filepath=None, initial_board=None):
        state.init()
        if initial_board:
            state.board = initial_board
        code = preprocess(text)
        old_output = sys.stdout
        redirected_output = sys.stdout = StringIO()

        final_board = None
        error_message = ''
        try:
            exec(code, builtins, builtins)
            final_board = state.board
        except Exception as e:
            error_message = format_exc(e).replace('<string>', filepath)
        sys.stdout = old_output
        result = GobspyResult(code, redirected_output.getvalue(), final_board, error_message)
        return result

Gobstones = Gobspy

run = lambda *args, **kwargs: Gobspy().run(*args, **kwargs)



from ...i18n import translate as t

class Enum(object):
    """Represents an enumerated type."""

    def __init__(self, i):
        self._ord = int(i)

    def enum_type(self):
        """Subclasses should implement the method to return the name
        of the enumerated type."""
        raise Exception("Subclass responsibility")

    def enum_size(self):
        """Subclasses should implement the method to return the number
        of elements in the enumerated type."""
        raise Exception("Subclass responsibility")

    def next(self):
        """Returns the next element in the enumerated type. (Wrap around
        if the maximum is reached)."""
        return self.__class__((self._ord + 1) % self.enum_size())

    def prev(self):
        """Returns the previous element in the enumerated type. (Wrap around
        if the minimum is reached)."""
        return self.__class__((self._ord - 1) % self.enum_size())

    def opposite(self):
        """Returns the opposite element in the enumerated type.
        Currently only works for enums of an even number of elements,
        returning the opposite element if they were in a circle."""
        new_i = (self._ord + self.enum_size() / 2) % self.enum_size()
        return self.__class__(new_i)

    def ord(self):
        """Returns the ord of the instance in the enumerated type."""
        return self._ord

    def __eq__(self, other):
        return isinstance(other, Enum) and \
               self.enum_type() == other.enum_type() and \
               self._ord == other.ord()


class NamedEnum(Enum):

    def name(self):
        pass

    def initial(self):
        return self.name()[0]

    def i18n_name(self):
        return t(self.name())

    def i18n_initial(self):
        return self.i18n_name()[0]



#
# Copyright (C) 2011-2017 Ary Pablo Batista <arypbatista@gmail.com>, Pablo Barenbaum <foones@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# # You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#

import os
import sys
import re
import colorsys
import webbrowser
import tempfile
import atexit
import shutil

def python_major_version():
    return sys.version_info[0]

if python_major_version() < 3:
    from StringIO import StringIO
else:
    from io import StringIO

try:
    import hashlib as md5
except ImportError:
    import md5

#### Various utility functions

VERSION = 0, 1, 1

def version_number():
    return '%s.%s.%s' % VERSION

class SourceException(Exception):
    def __init__(self, msg, area):
        self.msg = msg
        self.area = area
    def __repr__(self):
        s = ''
        if self.area:
            s += '\n%s\n' % (self.area,)
        s += '%s\n' % (indent(self.msg),)
        return s
    def error_type(self):
        return 'Error'

class StaticException(SourceException):
    pass

class DynamicException(SourceException):
    pass

def trim(x):
    return x.strip(' \t\r\n')

_blanks = re.compile('[ \t\r\n]+')
def trim_blanks(x):
    return trim(_blanks.sub(' ', x))

def nonempty(x):
    return x != ''

def read_file(fn):
    f = open(fn, 'r')
    c = f.read()
    f.close()
    return c

## set

def set_new(xs=[]):
    d = {}
    for x in xs: d[x] = 1
    return d

def set_add(s, x):
    s[x] = 1

def set_add_change(s, x):
    if x not in s:
        s[x] = 1
        return True
    return False

def set_extend(s1, s2):
    for k, v in s2.items():
        s1[k] = v

def set_remove(s1, s2):
    for k, v in s2.items():
        if k in s1:
            del s1[k]

## tokset
## Set with an associated token.
## When joining two sets, it always keeps
## the token that occurs first in the input.

tokset_key = lambda tree: tree.pos_begin.start

tokset_new = set_new

def tokset_new_from_dict(d):
    return d

def tokset_empty(s):
    return len(s) == 0

def tokset_extend_change(s1, s2):
    change = False
    for k, v in s2.items():
        if k not in s1:
            s1[k] = v
            change = True
        elif tokset_key(v) < tokset_key(s1[k]):
            s1[k] = v
    return change

tokset_extend = tokset_extend_change

def tokset_union(s1, s2):
    res = {}
    for k, v in s1.items():
        res[k] = v
    tokset_extend_change(res, s2)
    return res

def tokset_difference(s1, s2):
    res = {}
    for k, v in s1.items():
        if k not in s2:
            res[k] = v
    return res

def seq_sorted(xs, key=lambda x: x):
    if python_major_version() < 3:
        xs = [(key(x), x) for x in xs]
        xs.sort(lambda a, b: cmp(a[0], b[0]))
        return [xy[1] for xy in xs]
    else:
        return list(sorted(xs, key=key))

def seq_reversed(xs):
    ys = []
    for x in xs:
        ys.insert(0, x)
    return ys

def seq_no_repeats(xs):
    res = []
    for x in xs:
        if x not in res:
            res.append(x)
    return res

def seq_insert(xs, idx, ys):
    assert idx >= 0
    for y in ys:
        xs.insert(idx, y)
        idx += 1

def is_int(x):
    if not x:
        return False
    if not isinstance(x, str):
        return False
    if len(x) == 0:
        return False
    for c in x:
        if c not in '0123456789':
            return False
    return True

##

def dict_min_value(xs, key=lambda x: x):
    for x in seq_sorted(xs.values(), key):
        return x

def indent(msg, n=4):
    return '\n'.join([n * ' ' + m for m in msg.split('\n')])

def expand_tabs(x):
    return x.replace('\t', ' ')

def std_warn(x):
    sys.stderr.write(repr(x))

def show_string(s):
    s = s[1:-1]
    r = ''
    conv = {
        'a': '\a',
        'b': '\b',
        'f': '\f',
        'n': '\n',
        'r': '\r',
        't': '\t',
        'v': '\v',
        '\\': '\\',
        '\"': '\"'
    }
    i = 0
    while i < len(s):
        if s[i] == '\\' and i + 1 < len(s):
            c = s[i + 1]
            r += conv.get(c, c)
            i += 2
        else:
            r += s[i]
            i += 1
    return r

##

def md5sum(s):
    return md5.md5(s).hexdigest()

##

def hsv(h, s, v):
    r, g, b = colorsys.hsv_to_rgb(h, s, v)
    return '#%.2x%.2x%.2x' % (int(r * 255), int(g * 255), int(b * 255))

## Parser for option switches

def default_options(option_switches):
    opt = {}
    for o in option_switches:
        o = o.split(' ')
        sw = o[0][2:]
        if sw[:3] == 'no-':
            neg = True
            sw = sw[3:]
        else:
            neg = False
        if len(o) == 1:
            opt[sw] = neg
        else:
            opt[sw] = []
    return opt

def parse_options(option_switches, args, max_args=None):
    arguments = []
    opt = default_options(option_switches)
    i = 1
    n = len(args)
    while i < len(args):
        o = None
        for oi in option_switches:
            oi = oi.split(' ')
            if oi[0] == args[i]:
                o = oi
                break
        if o is None:
            if len(arguments) == max_args:
                return False
            arguments.append(args[i])
            i += 1
            continue

        sw = o[0][2:]
        if len(o) == 1:
            if sw[:3] == 'no-':
                neg = True
                sw = sw[3:]
            else:
                neg = False
            opt[sw] = not neg
            i += 1
        else:
            k = 1
            i += 1
            while k < len(o):
                if i >= n: return False
                opt[sw].append(args[i])
                i += 1
                k += 1
    return arguments, opt

##

TEMP_HTML_DIR = None

def temp_html_dir():
    global TEMP_HTML_DIR
    if TEMP_HTML_DIR is None:
        TEMP_HTML_DIR = tempfile.mkdtemp()
        atexit.register(_temp_html_dir_cleanup)
    return TEMP_HTML_DIR

def _temp_html_dir_cleanup():
    if TEMP_HTML_DIR is not None:
        shutil.rmtree(TEMP_HTML_DIR)

def open_html(fn):
    webbrowser.open(fn, autoraise=True)

def open_temp_html(prefix, contents):
    fn = os.path.join(temp_html_dir(), prefix + '.html')
    f = open(fn, 'w')
    f.write(contents)
    f.close()
    open_html(fn)

##

def read_stripped_lines(f):

    def next_line():
        while True:
            l = f.readline()
            if l == '' or l.strip(' \t\r\n') == '%%':
                return False
            l = l.strip(' \t\r\n').split('#')[0].strip(' \t\r\n')
            if l != '':
                return re.sub('[ \t]+', ' ', l)

    lines = []
    while True:
        l = next_line()
        if not l:
            break
        lines.append(l)
    return lines


# Reading one char

class _Getch:
    """Gets a single character from standard input.  Does not echo to the screen."""
    def __init__(self):
        try:
            self.impl = _GetchWindows()
        except ImportError:
            self.impl = _GetchUnix()

    def __call__(self): return self.impl()


class _GetchUnix:
    def __init__(self):
        import tty, sys

    def __call__(self):
        import sys, tty, termios
        fd = sys.stdin.fileno()
        old_settings = termios.tcgetattr(fd)
        try:
            tty.setraw(sys.stdin.fileno())
            ch = sys.stdin.read(1)
        finally:
            termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
        return ch


class _GetchWindows:
    def __init__(self):
        import msvcrt

    def __call__(self):
        import msvcrt
        return msvcrt.getch()


getch = _Getch()


def clear():
  os.system('cls' if os.name == 'nt' else 'clear')



import unittest
from lang.builtins import *


class TestColors(unittest.TestCase):

    def test_min_color(self):
        self.assertEqual(minColor(), Azul)

    def test_max_color(self):
        self.assertEqual(maxColor(), Verde)

    def test_next_color(self):
        colors = [Verde, Azul, Negro, Rojo, Verde]
        for i in range(len(colors) - 1):
            self.assertEqual(siguiente(colors[i]), colors[i+1])

    def test_prev_color(self):
        colors = [Verde, Azul, Negro, Rojo, Verde]
        for i in range(1, len(colors)):
            self.assertEqual(previo(colors[i]), colors[i-1])

    def test_opposite_color(self):
        colors = [Azul, Negro, Rojo, Verde]
        opposites = [Rojo, Verde, Azul, Negro]
        for i in range(len(colors)):
            self.assertEqual(opuesto(colors[i]), opposites[i])


if __name__ == '__main__':
    unittest.main()



#
# Copyright (C) 2011-2017 Ary Pablo Batista <arypbatista@gmail.com>, Pablo Barenbaum <foones@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#


from .fmt_gbt import GbtBoardFormat
from .fmt_gbb import GbbBoardFormat
from .fmt_html import HtmlBoardFormat
from .fmt_json import JsonBoardFormat

AvailableFormats = {
  'gbt': GbtBoardFormat,
  'gbb': GbbBoardFormat,
  'html': HtmlBoardFormat,
  'json': JsonBoardFormat
}

DefaultFormat = 'gbb'

def is_board_filename(filename):
  return filename.lower().split('.')[-1] in AvailableFormats

def format_for(filename):
  fmt = filename.lower().split('.')[-1]
  if fmt in AvailableFormats:
    return fmt
  else:
    return DefaultFormat



from .builtins import builtins as original_builtins
from ..i18n import translate as t

for k in original_builtins.keys():
    if k[0] != '_':
        try:
            value = original_builtins[k]
            code = '%s = value' % k
            code = '%s = value' % t(k)
            exec(code)
        except e:
            pass

builtins = locals()



from .enum import NamedEnum
from ...i18n import translate as t

#### Directions


DIRECTION_NAMES = [
    'North',
    'East',
    'South',
    'West',
]


DIRECTION_DELTA = {
    0: (1, 0),
    1: (0, 1),
    2: (-1, 0),
    3: (0, -1),
}


class Direction(NamedEnum):
    "Represents a Gobstones direction."

    def name(self):
        "Return the name of this direction."
        return DIRECTION_NAMES[self.ord()]

    def enum_type(self):
        "Return the name of the enumerated type."
        return 'Dir'

    def enum_size(self):
        "Return the size of the enumerated type."
        return 4

    def delta(self):
        "Return the delta for this direction."
        return DIRECTION_DELTA[self.ord()]

    def __repr__(self):
        return self.name()

DIRECTIONS = [Direction(i) for i in range(len(DIRECTION_NAMES))]

DIRECTIONS_BY_NAME = dict([(d.name(), d) for d in DIRECTIONS])

NUM_DIRECTIONS = len(DIRECTIONS)



from .exceptions import StaticException

import re

IMPORTS = [
    'from lang.builtins import *'
]

class PreprocessorException(StaticException):
    pass

class Preprocessor:

    def process(self, text):
        res = text
        self.check_root_commands(text)
        res = self.process_procedures(res)
        res = self.process_functions(res)
        res = self.process_main(res)
        return res

    def check_root_commands(self, text):
        root_commands = list(re.finditer(r'^[^\#dfp\s][^eur\s][^fno\s]', text, re.MULTILINE))
        if len(root_commands) > 0:
            raise PreprocessorException('You cannot call commands outside a procedure or a function definition')

    def process_procedures(self, text):
        text = text.replace(r"procedure\s+([a-zA-Z]+)", 'proc')
        return text.replace('proc ', 'def ')

    def process_main(self, text):
        return text + '\n'.join([
            '', '',
            '# Call Main procedure',
            'Main()'
        ])

    def process_functions(self, text):
        res = text + '\n\n# Define functions'
        res = res.replace(r"function\s+([a-zA-Z]+)", 'func')
        for match in re.finditer(r"func\s+([a-zA-Z]+)", text):
            name = match.group(1)
            res += '\n%s = define_function(%s)' % (name, name)
        res = res.replace('func ', 'def ')
        return res



preprocess = Preprocessor().process
